import numpy as np
import scipy as sp

import matplotlib as mpl
from matplotlib.figure import Figure
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qtagg import NavigationToolbar2QT as NavigationToolbar

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import *

from DNMR.tab import *

class TabFourierTransform(Tab):
    def __init__(self, data_widgets, parent=None):
        super(TabFourierTransform, self).__init__(data_widgets, 'tab_ft', parent)
        
        self.data = (np.array([]), np.array([]))
        self.fit_data = np.array([])
        self.left_pivot = 1
        self.right_pivot = -1
        self.moving_left = True

    def generate_layout(self):
        #self.button_fit = QPushButton('Fit Gaussian')
        #self.button_fit.clicked.connect(self.fit)
        
        self.label_integration_width = QLabel('Width (MHz)')
        self.spinbox_integration_width = QDoubleSpinBox()
        self.spinbox_integration_width.setRange(0, 1e9)
        self.spinbox_integration_width.setSingleStep(1e-6)
        self.spinbox_integration_width.setValue(0.8)
        self.spinbox_integration_width.setDecimals(6)
        self.spinbox_integration_width.valueChanged.connect(self.process_integrationwidth)
        
        self.label_integration_centre = QLabel('Centre (MHz)')
        self.spinbox_integration_centre = QDoubleSpinBox()
        self.spinbox_integration_centre.setRange(-1e9, 1e9)
        self.spinbox_integration_centre.setSingleStep(1e-6)
        self.spinbox_integration_centre.setValue(0.0)
        self.spinbox_integration_centre.setDecimals(6)
        self.spinbox_integration_centre.valueChanged.connect(self.process_integrationcentre)
        
        #self.canvas.mpl_connect('button_press_event', self.process_button)
        
        l = QVBoxLayout()
        l1 = QHBoxLayout()
        l1.addWidget(self.label_integration_width)
        l1.addWidget(self.spinbox_integration_width)
        l2 = QHBoxLayout()
        l2.addWidget(self.label_integration_centre)
        l2.addWidget(self.spinbox_integration_centre)
        l.addLayout(l1)
        l.addLayout(l2)
        return l
        
        #l = QHBoxLayout()
        #l.addWidget(self.button_fit)
        #return l

    def process_integrationwidth(self):
        self.spinbox_integration_width.setSingleStep(0.05 * self.spinbox_integration_width.value())
        self.update()
        
    def process_integrationcentre(self):
        self.update()

    def process_button(self, event):
        nav_state = self.ax.get_navigate_mode()
        if(nav_state is None):
            if(event.button == 1):
                if not(event.xdata is None):
                    if(self.moving_left):
                        self.left_pivot = event.xdata
                    else:
                        self.right_pivot = event.xdata
                    
                    self.moving_left = not(self.moving_left)
                    self.update()

    def plot_logic(self):
        index = self.fileselector.spinbox_index.value()

        times = self.data_widgets['tab_phase'].data[0]
        complexes = self.data_widgets['tab_phase'].data[1]

        timespacing = (times[index][1]-times[index][0])
    
        time_index = np.argmin(np.abs(self.fileselector.data['peak_locations'][:,None] - times), axis=1)
        s_complexes = np.zeros_like(complexes)
        s_times = np.zeros_like(times)
        for i in range(s_complexes.shape[0]):
            s_complexes[i,:] = np.roll(complexes[i,:], -time_index[i])
            s_times[i,:]     = np.roll(times[i,:],     -time_index[i])
        s_reals = np.real(s_complexes)
        s_imags = np.imag(s_complexes)

        fftfreq = np.fft.fftshift(np.fft.fftfreq(s_complexes[index].shape[0], d=timespacing)) # microseconds
        self.spinbox_integration_centre.setRange(np.min(fftfreq), np.max(fftfreq))
        
        self.left_pivot = self.spinbox_integration_centre.value() - self.spinbox_integration_width.value()/2.0
        self.right_pivot = self.spinbox_integration_centre.value() + self.spinbox_integration_width.value()/2.0
        
        fft = np.zeros_like(s_complexes, dtype=np.complex128)
        for i in range(s_complexes.shape[0]):
            fft[i] = np.fft.fftshift(np.fft.fft(s_complexes[i]))
        self.data = (fftfreq, fft)
        
        self.ax.plot(fftfreq, np.real(fft[index]), 'r', alpha=0.6, label='R')
        self.ax.plot(fftfreq, np.imag(fft[index]), 'b', alpha=0.6, label='I')
        self.ax.plot(fftfreq, np.abs(fft[index]), 'k', alpha=0.3, label='abs')
        
        if(self.fit_data.shape[0] > 0):
            self.ax.plot(fftfreq, self.fit_data, 'k--', alpha=0.5, label='fit')
        
        self.ax.set_xlabel('frequency (MHz)')

        self.ax.axvline(self.left_pivot, color='k')
        self.ax.axvline(self.right_pivot, color='k')

    def fit(self):
        '''Fits a gaussian'''
        def gauss(args, x):
            return args[0]*np.exp(-np.square((x-args[2])/(2*args[1]))) + args[3]
        
        res = sp.optimize.minimize(lambda args: np.sum(np.square(np.abs(self.data[1]) - gauss(args, self.data[0]))), 
                                   x0=[np.max(np.abs(self.data[1])),0.1,0.0,np.std(np.abs(self.data[1]))],
                                   bounds=[[0, np.max(np.abs(self.data[1]))*2],[0, np.max(self.data[0])/2.0],[0, np.max(self.data[0])/2.0],[0,np.max(np.abs(self.data[1]))*2]],
                                   method='Nelder-Mead')
        print(res)
        self.fit_data = gauss(res.x, self.data[0])
        self.update()
        
    def get_exported_data(self):
        index = self.fileselector.spinbox_index.value()
        return { 'times': self.data_widgets['tab_phase'].data[0][index],
                 'complexes': self.data_widgets['tab_phase'].data[1][index],
                 'frequencies (MHz)': self.data[0],
                 'fft': self.data[1][index],
               }



import sys
import pathlib
import traceback

import numpy as np
import scipy as sp

import matplotlib as mpl
from matplotlib.figure import Figure
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qtagg import NavigationToolbar2QT as NavigationToolbar

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import *
from PyQt6 import QtGui

import DNMR.fileops as fileops
from DNMR.miniwidgets import *

from DNMR.tab_phase_adj import *
from DNMR.tab_fourier_transform import *
from DNMR.tab_t1_fitting import *
from DNMR.tab_field_scan import *
from DNMR.tab_peak_amplitude import *
from DNMR.tab_inv_laplace import *

class MainWindow(QWidget):
    def __init__(self, parent=None):
        super(MainWindow, self).__init__(parent)
        path_to_icon = str(pathlib.Path(__file__).parent.absolute())+'/icon_transparent.png'
        pixmap = QtGui.QPixmap()
        pixmap.loadFromData(pathlib.Path(path_to_icon).read_bytes())
        appIcon = QtGui.QIcon(pixmap)
        
        self.setWindowIcon(appIcon)

        self.tabwidget_tabs = QTabWidget()
        data_widgets = {}
        self.fileselector = FileSelectionWidget()
        data_widgets['fileselector'] = self.fileselector
        if(len(sys.argv) > 1): # passed arguments are files to load
            self.fileselector.load_files(sys.argv[1:])
        
        self.pushbutton_process = QPushButton('Reload')
        self.pushbutton_process.clicked.connect(self.update_all)
        
        self.filedialog_export = QFileDialog()
        self.button_export = QPushButton('Export Data (CSV)')
        self.button_export.clicked.connect(self.export_selected)
        
        self.tab_phaseadj = TabPhaseAdjustment(data_widgets, self)
        self.tab_ft = TabFourierTransform(data_widgets, self)
        self.tab_t1 = TabT1Fit(data_widgets, self)
        self.tab_fieldscan = TabFieldScan(data_widgets, self)
        self.tab_peakamp = TabPeakAmplitude(data_widgets, self)
        self.tab_inv_laplace = TabInvLaplace(data_widgets, self)

        self.tabwidget_tabs.addTab(self.tab_phaseadj, 'Time Domain')
        self.tabwidget_tabs.addTab(self.tab_ft, 'Freq. Domain')
        self.tabwidget_tabs.addTab(self.tab_t1, 'T1 Fit')
        self.tabwidget_tabs.addTab(self.tab_fieldscan, 'Field Scan')
        self.tabwidget_tabs.addTab(self.tab_peakamp, 'Peak Amplitudes')
        self.tabwidget_tabs.addTab(self.tab_inv_laplace, 'Inverse Laplace')
        self.tabwidget_tabs.currentChanged.connect(lambda: self.tabwidget_tabs.currentWidget().update())

        layout = QVBoxLayout()
        layout.addWidget(self.tabwidget_tabs)
        layout.addWidget(self.fileselector)
        layouth = QHBoxLayout()
        layouth.addWidget(self.pushbutton_process)
        layouth.addWidget(self.button_export)
        layout.addLayout(layouth)
        self.setLayout(layout)

    def export_selected(self):
        fn = self.filedialog_export.getSaveFileName()[0]
        saved_dict = self.tabwidget_tabs.currentWidget().get_exported_data()
        print(saved_dict)
        df = pd.DataFrame(dict([ (k, pd.Series(v)) for k,v in saved_dict.items() ]))
        df.to_csv(path_or_buf=fn)
        print('Exporting dataframe:')
        print(df)

    def update_all(self):
        ct = self.tabwidget_tabs.count()
        for i in range(ct):
            self.tabwidget_tabs.widget(i).update()

def start_app():
    print('Starting QT. Please wait...')
    app = QApplication(sys.argv)
    main = MainWindow()
    main.setWindowTitle('DNMR')
    main.resize(640, 960)
    main.show()

    try:
        app.exec()
    except:
        traceback.print_exc
        
if __name__=='__main__':
    start_app()



import numpy as np
import scipy as sp

import matplotlib as mpl
from matplotlib.figure import Figure
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qtagg import NavigationToolbar2QT as NavigationToolbar

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import *

import DNMR.fileops as fileops
from DNMR.miniwidgets import *

from DNMR.tab import Tab

class TabT1Fit(Tab):
    output_frames = {}

    def __init__(self, data_widgets, parent=None):
        super(TabT1Fit, self).__init__(data_widgets, 'tab_t1_fitting', parent)
        
        self.data = (np.array([]), np.array([]))
        self.plot_data = (np.array([]), np.array([]))
        self.excluded_points_indices = []
        self.x0 = None
        self.sigmas = None

    def generate_layout(self):
        self.combobox_fittingroutine = QComboBox()
        self.combobox_fittingroutine.currentIndexChanged.connect(self.update_fit_type)
        
        self.pushbutton_fit = QPushButton('Fit')
        self.pushbutton_fit.clicked.connect(self.fit)
        
        self.checkbox_normalize = QCheckBox('Normalize?')

        l = QHBoxLayout()
        lv = QVBoxLayout()
        lv.addWidget(self.combobox_fittingroutine)
        lv.addWidget(self.checkbox_normalize)
        l.addLayout(lv)
        l.addWidget(self.pushbutton_fit)

        def add_fit_frame(name, *args):
            # fit output
            frm = QFrame() # TODO: Make this better.
            frm.hide()
            lo = QVBoxLayout()
            self.output_frames[name] = [ frm ]
            for i in range(len(args)//2):
                w = FitParameterWidget(args[2*i], args[2*i+1])
                #w = QLineEdit('fitting...')
                #fix = QCheckBox('Fix?')
                #li = QHBoxLayout()
                #li.addWidget(fix)
                #li.addWidget(w)
                lo.addWidget(w)
                self.output_frames[name] += [ {'widget': w } ]
            self.combobox_fittingroutine.addItem(name)
                
            frm.setLayout(lo)
            l.addWidget(frm)

        # Title, var_name, var_units, var_name, var_units, ...
            # DEVELOPER NOTE: If you want to add more options for this, make sure to define fit_func in ``fit`` below
        add_fit_frame('7/2 Spin',          '\u03b3\u2080', '', 's', '', 'T\u2081', '\u03bcs', 'r', '')
        add_fit_frame('7/2 Spin (Sat. 1)', '\u03b3\u2080', '', 's', '', 'T\u2081', '\u03bcs', 'r', '')
        add_fit_frame('1/2 Spin',          '\u03b3\u2080', '', 's', '', 'T\u2081', '\u03bcs', 'r', '')
        #add_fit_frame('Spin 1', '\u03b30', '', 's', '', 'T1', '\u03bcs', 'r', '')
        # ...
        
        self.update_fit_type()
        
        self.canvas.mpl_connect('button_press_event', self.process_button)

        return l

    def process_button(self, event):
        if(event.button == 1):
            if not(event.xdata is None):
                screenspace_data = self.ax.transData.transform(np.array([self.data[0], self.data[1]]).T).T
                screenspace_click = self.ax.transData.transform((event.xdata, event.ydata))
                
                xdist = np.square(screenspace_click[0] - screenspace_data[0])
                ydist = np.square(screenspace_click[1] - screenspace_data[1])
                selected_point_index = np.argmin(xdist + ydist)
                if(selected_point_index in self.excluded_points_indices):
                    self.excluded_points_indices.remove(selected_point_index)
                else:
                    self.excluded_points_indices += [selected_point_index]
                self.update()

    def plot_logic(self):
        freq = self.data_widgets['tab_ft'].data[0]
        ft   = self.data_widgets['tab_ft'].data[1]
        real = np.real(ft)
        try:
            del_times = self.fileselector.data.sequence['0'].delay_time
        except:
            del_times = self.fileselector.data.sequence['0'].relaxation_time # Legacy
            print(del_times)

        integrations = np.zeros(real.shape[0], dtype=np.complex128)
        start_index = np.argmin(np.abs(self.data_widgets['tab_ft'].left_pivot - freq))
        end_index = np.argmin(np.abs(self.data_widgets['tab_ft'].right_pivot - freq))
        if(end_index < start_index):
            tmp = start_index
            start_index = end_index
            end_index = tmp

        integrations = np.sum(real[:,start_index:end_index], axis=1)
        
        if(self.checkbox_normalize.isChecked()):
            integrations /= np.max(integrations)
        rt = np.real(self.data_widgets['tab_ft'].data[1])
        
        uncertainties = 1e-6*np.ones_like(integrations) # TODO: Figure out real stddevs
        #uncertainties += integrations * np.sqrt((end_index-start_index+1) / rt.shape[1])
        #uncertainties = np.abs(uncertainties)
            
        sort_indices = np.argsort(del_times)
        del_times = del_times[sort_indices]
        integrations = integrations[sort_indices]
        uncertainties = uncertainties[sort_indices]
        
        self.ax.set_xscale('log')
        self.ax.set_xlabel('delay time (us)')
        plotted_integrations = []
        plotted_del_times = []
        plotted_errs = []
        excluded_integrations = []
        excluded_del_times = []
        for i in range(len(integrations)):
            if not(i in self.excluded_points_indices):
                plotted_integrations += [integrations[i]]
                plotted_del_times += [del_times[i]]
                plotted_errs += [uncertainties[i]]
            else:
                excluded_integrations += [integrations[i]]
                excluded_del_times += [del_times[i]]
        plt_pts = self.ax.errorbar(plotted_del_times, plotted_integrations, label='integrations', linestyle='', marker='o', yerr=plotted_errs)
        self.ax.scatter(excluded_del_times, excluded_integrations, color=(plt_pts[-1][-1]).get_color(), linestyle='', marker='x')
        
        post_aq_max = np.max(self.fileselector.data.params.post_acquisition_time * 1e3) # this is in ms. Our axes in us
        self.ax.axvline(post_aq_max, linestyle='--', color='k')

        self.data = (del_times, integrations, uncertainties)

        if(self.plot_data[0].shape[0] > 0):
            params_list = ''
            out_frame = self.output_frames[self.combobox_fittingroutine.currentText()]
            for i in out_frame[1:]:
                params_list += f'{i['widget'].get_full_display()}\n'
            params_list = params_list[:-1]
            self.ax.plot(self.plot_data[0], self.plot_data[1], label=params_list)
        
    def update_fit_type(self):
        for key, val in self.output_frames.items():
            val[0].hide()
        out_frame = self.output_frames[self.combobox_fittingroutine.currentText()]
        out_frame[0].show()
        
    def fit(self):
        self.update() # get most recent values to fit
        self.plot_data = (np.array([]),np.array([]))
        out_frame = self.output_frames[self.combobox_fittingroutine.currentText()]
        bounds = None
        try:
            del_times = self.fileselector.data.sequence['0'].delay_time
        except:
            del_times = self.fileselector.data.sequence['0'].relaxation_time # Legacy, as I didn't know what this was when I wrote it. Surprise, surprise

        if(self.combobox_fittingroutine.currentText() == '7/2 Spin'):
            # DEVELOPER NOTE: If you want to add more options for this, make sure to define fit_func (similarly to below) and add an item in the generate_layout function
            
            bounds = [ [0, np.max(np.abs(self.data[1]))*10], [-1, 10], [np.min(del_times)/10, np.max(del_times)*10], [0.99*0, 1.01*10] ]
            def fit_func(args, x):
                gamma_0 = args[0]
                s = args[1] # inversion
                T1 = args[2] # relaxation time (actual fit variable, really)
                r = args[3] # stretched exponent (ideally 1)
                #y = y0 (1-(1+s) ((1/84)*Exp[-(t/T1)^r]+(3/44)*Exp[-(6 t/T1)^r]+(75/364)*Exp[-(15 t/T1)^r]+(1225/1716)*Exp[-(28 t/T1)^r]))
                fit = gamma_0 * (1-(1+s)*(
                                            (1/84)*     np.exp(-np.pow(x/T1,    r)) + 
                                            (3/44)*     np.exp(-np.pow(6*x/T1,  r)) +
                                            (75/364)*   np.exp(-np.pow(15*x/T1, r)) +
                                            (1225/1716)*np.exp(-np.pow(28*x/T1, r)) 
                                         ))
                return fit
                
        elif(self.combobox_fittingroutine.currentText() == '7/2 Spin (Sat. 1)'):
            bounds = [ [0, np.max(np.abs(self.data[1]))*10], [-1, 10], [np.min(del_times)/10, np.max(del_times)*10], [0.99*0, 1.01*10] ]
            
            def fit_func(args, t):
                gamma_0 = args[0]
                s = args[1]
                T1 = args[2]
                r = args[3]
                
                return gamma_0 * (1 - (1+s) * (1/84*np.exp(-np.pow(t/T1, r)) + 
                                               1/84*np.exp(-np.pow(3*t/T1, r)) + 
                                               2/66*np.exp(-np.pow(6*t/T1, r)) + 
                                               18/154*np.exp(-np.pow(10*t/T1, r)) + 
                                               1/1092*np.exp(-np.pow(15*t/T1, r)) + 
                                               49/132*np.exp(-np.pow(21*t/T1, r)) + 
                                               392/858*np.exp(-np.pow(28*t/T1, r))))
                
        elif(self.combobox_fittingroutine.currentText() == '1/2 Spin'):
            bounds = [ [0, np.max(np.abs(self.data[1]))*10], [-1, 10], [np.min(del_times)/10, np.max(del_times)*10], [0.99*0, 1.01*10] ]
            
            def fit_func(args, t):
                gamma_0 = args[0]
                s = args[1]
                T1 = args[2]
                r = args[3]
                
                return gamma_0 * (1 - (1+s) * np.exp(-np.pow(t/T1, r)))
            
        def cost_func(args, x, y, yerr):
            return np.sum(np.square((fit_func(args, x) - y)/np.maximum(yerr, 0.01))) # more points is more fits
            
        for i in range(len(out_frame)-1):
            if(out_frame[i+1]['widget'].is_fixed()):
                # Fix
                fv = out_frame[i+1]['widget'].get_value()
                bounds[i] = [ fv, fv ]
            
        included_xvals = []
        included_yvals = []
        included_errs = []
        for i in range(len(self.data[0])):
            if not(i in self.excluded_points_indices):
                included_xvals += [self.data[0][i]]
                included_yvals += [self.data[1][i]]
                included_errs  += [self.data[2][i]]
        included_xvals = np.array(included_xvals)
        included_yvals = np.array(included_yvals)
        included_errs  = np.array(included_errs)
        # global minimum
        res = sp.optimize.differential_evolution(lambda x: cost_func(x, 
                                                                     included_xvals, 
                                                                     included_yvals,
                                                                     included_errs), 
                                                 bounds=bounds)
        # get uncertainties on the fit, as I am too lazy to do the full analysis when scipy will do it for me
        try:
            picky_scipy_bounds = np.array(bounds).T 
            picky_scipy_bounds[0,:] -= 1e-9
            popt, pcov = sp.optimize.curve_fit(lambda xs, *args: fit_func(args, xs), included_xvals, included_yvals, p0=res.x, bounds=picky_scipy_bounds, sigma=included_errs, absolute_sigma=False)
        
            print(res)
            self.x0 = popt
            self.sigmas = np.sqrt(np.diag(pcov))
            x_vals = included_xvals
            if(x_vals.shape[0] < 100):
                x_vals = np.exp(np.linspace(np.log(np.min(x_vals*1e-1)), np.log(np.max(x_vals*1e1)), 100, endpoint=True))
            self.plot_data = (x_vals, fit_func(popt, x_vals))
            for i in range(len(self.x0)):
                try:
                    digits = int(np.ceil(np.abs(np.log10(self.sigmas[i]))))
                except:
                    digits = 10000 # sigma negative - a sign that something has gone horribly wrong and the user should deal with the drama. Show them the digits.
                if(self.sigmas[i] > 1.0):
                    rounded_digits = -digits+1
                else:
                    rounded_digits = digits
                    
                display_sigma = np.round(self.sigmas[i], rounded_digits)
                display_x = np.round(self.x0[i], rounded_digits)
                
                if not(out_frame[i+1]['widget'].is_fixed()):
                    out_frame[i+1]['widget'].set_value(display_x, display_sigma)
                #out_frame[i+1]['widget'].setText(f'{out_frame[i+1]["label"]}={display_x:,}\u00b1{display_sigma:,}{out_frame[i+1]["units"]}')
        except Exception as e:
            traceback.print_exc()
        self.update()
        
    def get_exported_data(self):
        out_frame = self.output_frames[self.combobox_fittingroutine.currentText()][1:]
        params_dict = {}
        if(self.x0 is not None):
            cnt = 0
            for i in out_frame:
                params_dict[i['label']] = [ str(self.x0[cnt]) + ' ' + i.units ]
                params_dict[i['label']+' error'] = [ str(self.sigmas[cnt]) + ' ' + i.units ]
                cnt += 1
        
        index = self.fileselector.spinbox_index.value()
        pd = {
                 'frequencies (MHz)': self.data_widgets['tab_ft'].data[0],
                 'fft': self.data_widgets['tab_ft'].data[1][index],
                 'delays': self.data[0],
                 'integrals': self.data[1],
                }
        pd.update(params_dict)
        return pd


import traceback

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import *
from PyQt6.QtGui import QDoubleValidator
import numpy as np

import DNMR.fileops as fileops

class FitParameterWidget(QWidget):
    def __init__(self, label, units, parent=None):
        super(FitParameterWidget, self).__init__(parent)
        
        self.label = label
        self.units = units
        
        self.checkbox_fix    = QCheckBox('Fix?')
        self.checkbox_fix.stateChanged.connect(self.update_fixed)
        
        self.label_parameter = QLabel(f'{self.label}=')
        self.label_units = QLabel(f'{self.units}')
        self.lineedit_value  = QLineEdit(f'?')
        self.lineedit_value.setValidator(QDoubleValidator())
        self.lineedit_value.setEnabled(False)
        
        li  = QHBoxLayout()
        li.addWidget(self.checkbox_fix)
        li.addWidget(self.label_parameter)
        li.addWidget(self.lineedit_value)
        li.addWidget(self.label_units)
        
        self.setLayout(li)
    
    def update_fixed(self):
        self.lineedit_value.setEnabled(self.is_fixed())
        
    def is_fixed(self):
        return self.checkbox_fix.isChecked()
        
    def get_value(self, units=False):
        if(units):
            return str(self.get_value() + self.units)
        try:
            return float(self.lineedit_value.text().split('\u00b1')[0])
        except:
            return 0
        
    def get_error(self, units=False):
        if(units):
            return str(self.get_error() + self.units)
        try:
            return float(self.lineedit_value.text().split('\u00b1')[1])
        except:
            return 0
        
    def set_value(self, v, uncertainty):
        self.lineedit_value.setText(f'{v}\u00b1{uncertainty}')
        
    def get_full_display(self):
        return f'{self.label}={self.lineedit_value.text()} {self.units}'

class PhaseAdjustmentWidget(QWidget):
    def __init__(self, parent=None, callback=lambda: None):
        super(PhaseAdjustmentWidget, self).__init__(parent)

        self.slider_phase = QSlider()
        self.slider_phase.setRange(-180, 180)
        self.slider_phase.valueChanged.connect(callback)
        self.slider_phase.valueChanged.connect(lambda v: self.label_phase.setText(f'Phase: {v}\u00b0'))
        self.slider_phase.setOrientation(Qt.Orientation.Horizontal)
        self.label_phase = QLabel('Phase: 0\u00b0')

        layout = QVBoxLayout()
        layout.addWidget(self.label_phase)
        layout.addWidget(self.slider_phase)
        self.setLayout(layout)

class FileInfoWidget(QWidget):
    def __init__(self, parent=None):
        super(FileInfoWidget, self).__init__(parent)
        
        self.listview_docinfo = QListWidget()
        
        layout = QVBoxLayout()
        layout.addWidget(self.listview_docinfo)
        self.setLayout(layout)
        
    def update_items(self, d, length=None, prefix=''):
        '''Takes a data_struct'''
        if(length is None):
            length = d['size']
        for i in list(d.keys()):
            if(isinstance(d[i], fileops.data_struct)):
                self.update_items(d[i], length=length, prefix=prefix+str(i)+'/')
            elif(isinstance(d[i], np.ndarray)):
                if(d[i].ndim == 1):
                    s = '\n'.join([ f'\t{j}: ' + str(d[i][j]) for j in range(length) ])
                    self.listview_docinfo.addItem(f'{prefix+i} (array, len={d[i].shape[0]})='+'{\n'+s+'\n}')
                elif(d[i].ndim == 2):
                    # first index is scan index, second is datapoint
                    s = '\n'.join([ f'\t{j}: ' + str(d[i][j]) for j in range(length) ])
                    self.listview_docinfo.addItem(f'{prefix+i} (array, len={d[i].shape[0]}x{d[i].shape[1]})='+'{\n'+s+'\n}')
            else:
                self.listview_docinfo.addItem(f'{prefix+i}={d[i]}')
                
class QuickInfoWidget(QWidget):
    def __init__(self, parent=None):
        super(QuickInfoWidget, self).__init__(parent)
        
        self.label_filetitle = QLabel('Current file: N/A')
        self.listview_envinfo = QListWidget()
        
        layout = QVBoxLayout()
        layout.addWidget(self.label_filetitle)
        layout.addWidget(self.listview_envinfo)
        self.setLayout(layout)
        
    def update_items(self, fns, d, index):
        self.listview_envinfo.clear()
        
        try:
            fmt_fns = [ f.split('/')[-1].split('\\')[-1] for f in fns ]
            self.label_filetitle.setText(f'Current file: {fmt_fns[0]}' if len(fmt_fns)==1 else f'Current files: {fmt_fns[0]} + ...')
            
            if('size' in d.keys()):
                self._update_items(d, index)
        except:
            self.label_filetitle.setText('Current file: N/A')
        
    def _update_items(self, d, index, length=None, prefix=''):
        '''Takes a data_struct. Updates with all keys starting with environment_'''
        if(length is None):
            length = d['size']
            
        for i in list(d.keys()):
            if(i[:len('environment_')] != 'environment_'):
                continue
            name = prefix+str(i[len('environment_'):])
            if(isinstance(d[i], fileops.data_struct)):
                self._update_items(d[i], index, length=length, prefix=name+'/')
            elif(isinstance(d[i], np.ndarray)):
                if(d[i].ndim == 1):
                    s = str(d[i][index])
                    self.listview_envinfo.addItem(f'{name}='+s)
                elif(d[i].ndim == 2):
                    # first index is scan index, second is datapoint
                    s = str(d[i][index])
                    self.listview_envinfo.addItem(f'{name}='+s)
            else:
                self.listview_envinfo.addItem(f'{name}={d[i]}')
        
        if('comments' in d.keys()):
            s = str(d['comments'][index][0].decode('utf-8'))
            self.listview_envinfo.addItem(f'comments: '+s)
        if('sample' in d.keys()):
            s = str(d['sample'][index][0].decode('utf-8'))
            self.listview_envinfo.addItem(f'sample: '+s)
        if('nucleus' in d.keys()):
            s = str(d['nucleus'][index][0].decode('utf-8'))
            self.listview_envinfo.addItem(f'nucleus: '+s)

class SequenceWidget(QWidget):
    def __init__(self, parent=None):
        super(SequenceWidget, self).__init__(parent)
        
        self.listview_seq = QListWidget()
        
        layout = QVBoxLayout()
        layout.addWidget(self.listview_seq)
        self.setLayout(layout)
        
    def update_items(self, d, index):
        self.listview_seq.clear()
        
        if('sequence' in d.keys()):
            self._update_items(d['sequence'], index)
        
    def _update_items(self, d, index):
        '''Takes a data_struct. Updates sequence visualisation'''
        pulse_strs = []
        pulse_indices = []
        
        max_width_pw = 0
        max_width_ph = 0
        max_width_pc = 0
        max_width_dt = 0
        max_width_pi = 0
        
        for i in list(d.keys()): # '0', '1', '2', etc. (pulses)
            try:
                pindex = int(i) # might have 'size', etc.
                if(len(str(pindex)) > max_width_pi):
                    max_width_pi = len(str(pindex))
            except:
                continue
                
            pw = str(d[i]['pulse_width'][index])
            dt = str(d[i]['delay_time'][index])
            pc = str(d[i]['phase_cycle'][index])
            ph = str(d[i]['pulse_height'][index])
            
            if(len(pw) > max_width_pw):
                max_width_pw = len(pw)
            if(len(dt) > max_width_dt):
                max_width_dt = len(dt)
            if(len(pc) > max_width_pc):
                max_width_pc = len(pc)
            if(len(ph) > max_width_ph):
                max_width_ph = len(ph)
        
        for i in list(d.keys()): # '0', '1', '2', etc. (pulses)
            try:
                pindex = int(i) # might have 'size', etc.
            except:
                continue
                
            pw = d[i]['pulse_width'][index]
            dt = d[i]['delay_time'][index]
            pc = d[i]['phase_cycle'][index]
            ph = d[i]['pulse_height'][index]
            
            S = f'{pindex:>0{max_width_pi}}: {pw:>{max_width_pw}}\u03bcs @ {ph:>{max_width_ph}}% & {dt:>{max_width_dt}}\u03bcs delay'
            pulse_indices += [pindex]
            pulse_strs += [S]
        
        pulse_strs = np.array(pulse_strs)[np.argsort(np.array(pulse_indices))]
            
        for i in pulse_strs:
            self.listview_seq.addItem(i)

class FileSelectionWidget(QWidget):
    def __init__(self, parent=None):
        super(FileSelectionWidget, self).__init__(parent)
        
        self.filedialog = QFileDialog()
        self.button_load = QPushButton('Load')
        self.button_load.clicked.connect(self.open_file)
        self.button_info = QPushButton('Info')
        self.button_info.clicked.connect(self.file_info)
        self.label_channel = QLabel('Channel:')
        self.spinbox_channel = QSpinBox()
        self.label_index = QLabel('Index:')
        self.spinbox_index = QSpinBox()
        self.checkbox_holdplots = QCheckBox('Hold plots')
        self.quickinfo_envinronment = QuickInfoWidget()
        self.sequence_info = SequenceWidget()

        layout = QGridLayout()
        l0 = QVBoxLayout()
        l0.addWidget(self.button_load)
        l0.addWidget(self.button_info)
        l = QHBoxLayout()
        l.addWidget(self.label_index)
        l.addWidget(self.spinbox_index)
        l2 = QHBoxLayout()
        l2.addWidget(self.label_channel)
        l2.addWidget(self.spinbox_channel)
        l0.addLayout(l)
        l0.addLayout(l2)
        l0.addWidget(self.checkbox_holdplots)
        l_info = QHBoxLayout()
        self.quickinfo_envinronment.setSizePolicy(QSizePolicy.Policy.Maximum,QSizePolicy.Policy.Maximum)
        l_info.addWidget(self.quickinfo_envinronment)
        self.sequence_info.setSizePolicy(QSizePolicy.Policy.Maximum,QSizePolicy.Policy.Maximum)
        l_info.addWidget(self.sequence_info)
        
        layout.addLayout(l0, 0, 0)
        layout.setColumnStretch(0,1)
        layout.addLayout(l_info, 0, 1)
        layout.setColumnStretch(1,2)
        
        self.setLayout(layout)

        self.fn = []
        self._fn = [[]] # for all channels
        self.data = {}
        self._data = [{}] # for all channels
        
        self.infodialogs = []

        self.callbacks = [ lambda: self.quickinfo_envinronment.update_items(self.fn, self.data, self.spinbox_index.value()) ]
        self.callbacks += [ lambda: self.sequence_info.update_items(self.data, self.spinbox_index.value()) ]
        self.spinbox_index.valueChanged.connect(self.callback)
        self.spinbox_channel.valueChanged.connect(self.channel_callback)
    
    def channel_callback(self):
        while(len(self._fn) <= self.spinbox_channel.value()):
            self._fn += [[]]
            self._data += [fileops.data_struct()]
        self.fn = self._fn[self.spinbox_channel.value()]
        self.data = self._data[self.spinbox_channel.value()]
        if(len(self.fn) > 0):
            self.spinbox_index.setRange(0, self.data['size']-1)
            self.label_index.setText(f'Index (/{self.data["size"]-1}, 0-indexed):')
        self.spinbox_index.setValue(0)
        self.callback()
    
    def callback(self):
        for i in self.callbacks:
            i()

    def load_files(self, fns):
        try:
            newch = self.spinbox_channel.value()
            try: # just in case the channel hasn't been made yet.
                while(len(self._fn[newch]) > 0):
                    newch += 1
            except:
                pass # we found an empty spot!
            big_data = fileops.get_data(fns[0])
            if(len(fns) > 1):
                for fn in fns[1:]:
                    data = fileops.get_data(fn)
                    big_data = big_data + data
            self.spinbox_channel.setValue(newch)
            self.fn = fns # above lines will throw exceptions if anything bad happens, so if anything bad happens, we want to preserve the previous file being loaded
            self.data = big_data
            self._fn[newch] = self.fn
            self._data[newch] = self.data
            self.spinbox_index.setRange(0, self.data['size']-1)
            self.label_index.setText(f'Index (/{self.data["size"]-1}, 0-indexed):')
            self.spinbox_index.setValue(0)
            self.callback()
        except:
            traceback.print_exc()

    def open_file(self):
        try:
            fns = self.filedialog.getOpenFileNames()[0]
        except Exception as e:
            traceback.print_exc()
        self.load_files(fns)
            
    def file_info(self):
        if(len(list(self.data.keys())) != 0):
            self.infodialogs += [FileInfoWidget()] # must be stored somewhere, or python garbage collection will clean it up and close the window
            self.infodialogs[-1].update_items(self.data)
            current_fns = self.fn
            filename_str = f'Info on file {current_fns[0]}' if len(current_fns) == 1 else f'Info on files {current_fns}'
            self.infodialogs[-1].setWindowTitle(filename_str)
            self.infodialogs[-1].show()
    


import numpy as np
import h5py as hdf
import traceback
import re

from DNMR.fileops_loaders.data_struct import data_struct, hdf_to_dict

def read_hdf_v100(file):
    toplevel = file.keys()
    points = []
    point_indices = []
    point_numbers = []
    count = 0
    for i in toplevel: # get all points
        m = re.match('entry(?P<index>[0-9]+)', i)
        if not(m is None):
            points += [ m[0] ]
            point_numbers += [ int(m['index']) ]
            point_indices += [ count ]
            count += 1
    points = np.array(points)
    point_indices = np.array(point_indices)
    point_numbers = np.array(point_numbers)
    
    sorted_indices = np.argsort(point_numbers)
    points = points[sorted_indices]
    point_numbers = point_numbers[sorted_indices]
    
    data = data_struct()
    special_group_keys = [ 'detectors', 'environment']
    
    def strip_key(k):
        if(k[:5] == 'tnmr_'):
            return k[5:]
        return k
        
    def parse_dataset_into_struct(parent_group, key, struct, placement_index, key_to_write=None):
        '''Takes in an HDF group (parent_group), a _dataset_ value's key, and writes it into the appropriate place in data_struct (struct) at placement_index.'''
        kstrip = strip_key(key)
        val = parent_group[key]
        
        #handle non-array data.
        try:
            if(val.ndim == 0):
                if(isinstance(val.dtype, np.dtypes.BytesDType)):
                    conv = np.array(val, 'S').tobytes().decode('utf-8') # for strings. See HDF docs.
                    val = [conv]
                else:
                    val = [val]
        except: # in the case of dictionaries/further groups.
            pass
        
        struct[(key if key_to_write is None else key_to_write)][placement_index] = val
    
    def get_formatted_key(k):
        ks = k
        group = ''
        if('/' in ks):
            group = ks.split('/')[0]
            ks = '/'.join(ks.split('/')[1:])
        ks = strip_key(ks)
        
        for i in special_group_keys:
            if i in group:
                return i+'_'+ks
        return ks
    
    # load the first one, to get sizes etc.
    for ikey, ival in file[points[0]].items():
        if(isinstance(ival, hdf.Dataset)):
            kstrip = strip_key(ikey)
            data[kstrip] = [ [0] ] * len(points)
        elif(isinstance(ival, hdf.Group)):
            for key, val in ival.items():
                kstrip = strip_key(key)
                data[get_formatted_key(ikey+'/'+key)] = [ [0] ] * len(points)
    
    data['size'] = len(points)
    
    # Grab all data in entries, write it into data struct.
    for i, index in zip(points, point_indices):
        for ikey, ival in file[i].items():
            if(isinstance(ival, hdf.Dataset)):
                parse_dataset_into_struct(file[i], ikey, data, index)
            elif(isinstance(ival, hdf.Group)):
                for key, val in ival.items():
                    parse_dataset_into_struct(ival, key, data, index, key_to_write=get_formatted_key(ikey + '/' + key))
        
    for key, val in data.items():
        # check if we can turn it into a dict, then numpy array
        try:
            ds = data_struct(hdf_to_dict(val[0]))
            for i in range(1, len(val)):
                ds += data_struct(hdf_to_dict(val[i]))
            data[key] = ds
        except:
            try:
                arr = np.array(val)
                try:
                    arr_f = arr.astype(float)
                    arr_i = arr.astype(int)
                    if(arr_i == arr_f):
                        arr = arr_i
                    else:
                        arr = arr_f
                except:
                    pass
                data[key] = arr
            except:
                pass
    
    if('actual_num_acqs' in data.params.keys()):
        print('Normalising signals from number of acquisitions')
        data['reals'] /= data.params.actual_num_acqs[:,None]
        data['imags'] /= data.params.actual_num_acqs[:,None]
                
    #print('#' * 100)
    #print(data)
    
    return data



import numpy as np
import scipy as sp

import matplotlib as mpl
from matplotlib.figure import Figure
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qtagg import NavigationToolbar2QT as NavigationToolbar

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import *

from DNMR.miniwidgets import *
from DNMR.tab import Tab
from DNMR.fileops import data_struct

class TabPeakAmplitude(Tab):
    def __init__(self, data_widgets, parent=None):
        super(TabPeakAmplitude, self).__init__(data_widgets, 'tab_peakamp', parent)
        
        self.fileselector.callbacks += [ self.retrieve_labels ]
        self.data = ([], [])
        
    def generate_layout(self):
        self.combobox_labelling = QComboBox()
        self.combobox_labelling.addItem('Load Order')
        self.combobox_labelling.setCurrentIndex(0)
        
        layout = QHBoxLayout()
        layout.addWidget(self.combobox_labelling)
        return layout
        
    def deconstruct_dict(self, d, plunge=False):
        keys = []
        for k in d.keys():
            try:
                if(isinstance(d[k], data_struct)):
                    ks = self.deconstruct_dict(d[k], True)
                    for i in range(len(ks)):
                        ks[i] = f'{k}/{ks[i]}'
                    keys += ks
                elif(plunge):
                    keys += [k]
                    raise Exception
                a = d[k][0] # iterable
                if(len(d[k]) == self.fileselector.data['size']): # good size
                    good = False
                    try:
                        a = a[0]  # not >1D
                        if(len(d[k][0]) == 1):
                            good = True # only pseudo->1D
                    except:
                        good = True
                    if(good):
                        if(isinstance(d[k][0], data_struct)):
                            ks = self.deconstruct_dict(d[k][0], True)
                            for i in range(len(ks)):
                                ks[i] = f'{k}/{ks[i]}'
                            keys += ks
                        else:
                            keys += [k]
            except:
                pass
        return keys
        
    def retrieve_labels(self):
        current_item = self.combobox_labelling.currentText()
        
        for i in range(self.combobox_labelling.count()):
            self.combobox_labelling.removeItem(0)
            
        self.combobox_labelling.addItem('Load Order')
        keys = self.deconstruct_dict(self.fileselector.data)
        for k in keys:
            self.combobox_labelling.addItem(k)
            
        if(current_item in ['Load Order'] + keys):
            self.combobox_labelling.setCurrentIndex((['Load Order'] + keys).index(current_item))

    def plot_logic(self):
        
        freq = self.data_widgets['tab_ft'].data[0]
        ft   = self.data_widgets['tab_ft'].data[1]
        real = np.real(ft)
        try:
            del_times = self.fileselector.data.sequence['0'].delay_time
        except:
            del_times = self.fileselector.data.sequence['0'].relaxation_time # Legacy

        integrations = np.zeros(real.shape[0])
        start_index = np.argmin(np.abs(self.data_widgets['tab_ft'].left_pivot - freq))
        end_index = np.argmin(np.abs(self.data_widgets['tab_ft'].right_pivot - freq))
        if(end_index < start_index):
            tmp = start_index
            start_index = end_index
            end_index = tmp

        integrations = np.sum(ft[:,start_index:end_index], axis=1)
        #integrations /= np.max(np.abs(integrations))
        integrals = integrations
        
        peaks = ft[:,len(ft)//2]
        integrals = peaks

        if(self.combobox_labelling.currentText() == 'Load Order'):
            indices = np.linspace(0, self.fileselector.data['size'], self.fileselector.data['size'], endpoint=False)
        else:
            k = self.combobox_labelling.currentText()
            ks = k.split('/')
            d0 = self.fileselector.data[ks[0]]
            for K in ks[1:]:
                d0 = d0[K]
            indices = d0
            indices = np.reshape(indices, shape=integrals.shape)
        self.ax.set_xlabel(self.combobox_labelling.currentText())
        sorter = np.argsort(indices)
        indices = indices[sorter]

        reals = np.real(integrals)
        imags = np.imag(integrals)
        mags = np.abs(integrals)
        
        self.ax.plot(indices, mags[sorter], alpha=0.6, label=f'Mag.', linestyle='none', marker='o')
        #self.ax.plot(indices, reals[sorter], 'r', alpha=0.6, label='R', linestyle='none', marker='o')
        #self.ax.plot(indices, imags[sorter], 'b', alpha=0.6, label='I', linestyle='none', marker='o')
        
        self.data = (indices, integrals)

    def get_exported_data(self):
        index = self.fileselector.spinbox_index.value()
        return { 'frequencies (MHz)': self.data_widgets['tab_ft'].data[0][index],
                 'fft': self.data_widgets['tab_ft'].data[1][index],
                 self.combobox_labelling.currentText(): self.data[0],
                 'integrals': self.data[1],
               }